﻿///////////////////////////////////////////////////////////////////////////////
//
// (C) 2017 ICE TEA GROUP LLC - ALL RIGHTS RESERVED
//
// 
//
// ALL INFORMATION CONTAINED HEREIN IS, AND REMAINS
// THE PROPERTY OF ICE TEA GROUP LLC AND ITS SUPPLIERS, IF ANY.
// THE INTELLECTUAL PROPERTY AND TECHNICAL CONCEPTS CONTAINED
// HEREIN ARE PROPRIETARY TO ICE TEA GROUP LLC AND ITS SUPPLIERS
// AND MAY BE COVERED BY U.S. AND FOREIGN PATENTS, PATENT IN PROCESS, AND
// ARE PROTECTED BY TRADE SECRET OR COPYRIGHT LAW.
//
// DISSEMINATION OF THIS INFORMATION OR REPRODUCTION OF THIS MATERIAL
// IS STRICTLY FORBIDDEN UNLESS PRIOR WRITTEN PERMISSION IS OBTAINED
// FROM ICE TEA GROUP LLC.
//
///////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Web;
using Wisej.Core;

namespace Wisej.Web.Ext.Uploader
{
	/// <summary>
	/// Xterm is a lightweight, free, and open source terminal viewer for the web.
	/// Built using HTML5, CSS3, JavaScript, and WebGL. See https://xtermjs.org/.
	/// </summary>
	[ToolboxBitmapAttribute(typeof(Uploader))]
	[Description("Input is a lightweight, free, and open source terminal for the web.")]
	public class Uploader : Widget, IWisejControl, IWisejComponent
    {
        private object myLock = new object();
        public event CancelEventHandler FileOk;
        public event EventHandler OnCancelled;
        public event EventHandler OnStarted;
        public event EventHandler<UploaderProgressEventArgs> UploadProgress;
        public event EventHandler<UploaderErrorEventArgs> Error;
        public event EventHandler OnFileReady;             
        private SpeedMonitor UploadMonitor = new SpeedMonitor();
        private bool IsRunning;
        public Uploader()
        {
            if (!string.IsNullOrEmpty(Application.StartupPath))
            {
                this.TempPath = Path.Combine(Application.StartupPath, "FileUploadTemp");
            }
        }

        public void BrowseForFile()
        {
            this.Call("selectFile");
        }

        [EditorBrowsable(EditorBrowsableState.Never)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public UploaderFile File
        {
            get
            {
                return Files.Single();
            }
        }

        [EditorBrowsable(EditorBrowsableState.Never)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public List<UploaderFile> Files { get; set; } = new List<UploaderFile>();

        public UploadFileInfo UploadFile { get; set; }

        #region Properties

        /// <summary>
        /// Overridden to create our initialization script.
        /// </summary>
        [Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public override string InitScript
		{
			get { return BuildInitScript(); }
			set { }
		}

        /// <summary>
        /// Overridden to return our list of script resources.
        /// </summary>
        [Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public override List<Package> Packages
		{
			get
			{
				if (base.Packages.Count == 0)
				{
                    // initialize the loader with the required libraries.
                    //base.Packages.Add(new Package()
                    //{
                    //    Name = "radar.css",
                    //    Source = GetResourceURL("Wisej.Web.Ext.Rad.JavaScript.radar.css")                        
                    //});
                }                

                return base.Packages;
			}
		}

		#endregion

		#region Wisej Implementation

		private string BuildInitScript()
		{
			IWisejControl me = this;
			dynamic options = new DynamicObject();
            string script = GetResourceString("Wisej.Web.Ext.Uploader.JavaScript.startup.js");
            return script;
		}

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Browsable(false)]
        public dynamic Options
        {
            get
            {
                dynamic options = new DynamicObject();  
                return options;
            }
        }

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
        [Browsable(true)]
        public Color MyBorderColor { get; set; }
        public string RadarImageUrl { get; private set; }
        public string TempPath { get; }
        public string FileName { get; private set; }
        public string FilePath { get; private set; }
        public long FileSize { get; private set; }
        public long BytesDone { get; private set; }
        public string PartialFilePath { get; private set; }

        private FileStream fileStream;

        protected override void OnWidgetEvent(WidgetEventArgs e)
        {
            base.OnWidgetEvent(e);
        }

        protected override void OnWebRender(dynamic config)
        {
            base.OnWebRender((object)config);
            IWisejComponent me = this;

            // render our new properties.
            config.borderStyle = this.BorderStyle;
            config.options = this.Options;

            config.webMethods = new[] { "UploadFileChunk", "UploadFileError", "UploadFileSuccess", "UploadFileStarted", "UploadFileCancelled", "UploadFileChanged" };           
        }

        #endregion

        private void InitializeComponent()
        {
            this.SuspendLayout();
            // 
            // Uploader
            // 
            this.Name = "Uploader";
            this.Size = new System.Drawing.Size(180, 36);
            this.ResumeLayout(false);
        }

        [WebMethod]
        public void UploadFileChunk(string base64, long offSet, int length, int partCount)
        {           
            Application.StartTask(() =>
            {
                var buffer = Convert.FromBase64String(base64);

                lock (myLock)
                {
                    if (this.fileStream.Length < offSet)
                    {
                        this.fileStream.SetLength(offSet);
                    }
                    this.fileStream.Position = offSet;
                    this.fileStream.Write(buffer, 0, length);
                    BytesDone += length;
                    UploadMonitor.AddDelta(length);
                }
            });
        }

        [WebMethod]
        public void UploadFileError(string error)
        {
            Error?.Invoke(this, new UploaderErrorEventArgs(UploadErrorType.Upload, error, new string[] { FileName }, new long[] { FileSize }));
        }

        [WebMethod]
        public void UploadFileSuccess()
        {
            IsRunning = false;
            this.fileStream.Flush();
            this.fileStream?.Dispose();
            var file = new UploaderFile()
            {
                FilePath = PartialFilePath
            };
            Files.Add(file);

            RunInTask(() =>
            {
                UpdateProgress();
                CancelEventArgs args = new CancelEventArgs();
                this.FileOk?.Invoke(this, args);
                if (args.Cancel)
                {
                    file.Delete();
                }
            });
        }

        public void RunInTask(Action action)
        {
            Application.RunInContext(this, () =>
            {
                Application.StartTask(() =>
                {
                    action.Invoke();
                    Application.Update(this);
                });
            });
        }

        [WebMethod]
        public void UploadFileStarted(string fileName, long fileSize)
        {
            BytesDone = 0;

            if (!Directory.Exists(TempPath))
            {
                Directory.CreateDirectory(TempPath);
            }

            this.FileName = fileName;
            this.FileSize = fileSize;
            this.PartialFilePath = $"{this.FilePath}.{DateTime.Now.Ticks}.partial";
            this.fileStream = new FileStream(this.PartialFilePath, FileMode.Create, FileAccess.Write, FileShare.ReadWrite | FileShare.Delete);
            IsRunning = true;
            _ = MonitorUpload();

            RunInTask(() => OnStarted?.Invoke(this, EventArgs.Empty));
        }

        [WebMethod]
        public void UploadFileCancelled()
        {
            IsRunning = false;
            RunInTask(() => OnCancelled?.Invoke(this, EventArgs.Empty));
        }

        [WebMethod]
        public void UploadFileChanged(string fileName, long fileSize, string type)
        {
            if (fileName == null)
            {
                this.UploadFile = null;
                RunInTask(() => OnFileReady.Invoke(this, EventArgs.Empty));
                return;
            }
            this.UploadFile = new UploadFileInfo()
            {
                FileName = fileName,
                FileSize = fileSize,
                Type = type,
            };
            RunInTask(() => OnFileReady.Invoke(this, EventArgs.Empty));
        }

        public class UploadFileInfo
        {
            public string FileName { get; internal set; }
            public long FileSize { get; internal set; }
            public string Type { get; internal set; }
        }

        public async Task MonitorUpload()
        {
            while (!this.IsDisposed && IsRunning)
            {
                if (this.UploadProgress != null)
                {
                    Application.Update(this, () =>
                    {
                        UpdateProgress();
                    });
                }

                await Task.Delay(1000);
            }
        }

        private void UpdateProgress()
        {
            UploadMonitor.Tick();
            this.UploadProgress?.Invoke(this, new UploaderProgressEventArgs(BytesDone, this.FileSize, UploadMonitor.Rate));
        }

        public void Cancel()
        {
            this.Call("cancelDownload");
        }

        public void StartUpload()
        {
            this.Call("startUpload");
        }
    }

    public class UploaderErrorEventArgs : EventArgs
    {
        //
        // Summary:
        //     Initializes a new instance of the Wisej.Web.UploadErrorEventArgs class with the
        //     specified error message.
        //
        // Parameters:
        //   type:
        //     The error type, one of Wisej.Web.UploadErrorType.
        //
        //   message:
        //     The error message.
        //
        //   fileNames:
        //     The list of files that failed to upload.
        //
        //   fileSizes:
        //     The list of files that failed to upload.
        public UploaderErrorEventArgs(UploadErrorType type, string message, string[] fileNames, long[] fileSizes)
        {
            this.ErrorType = type;
            this.Message = message;
            this.FileNames = fileNames;
            this.FileSizes = fileSizes;
        }

        //
        // Summary:
        //     Returns the error type.
        public UploadErrorType ErrorType { get; }
        //
        // Summary:
        //     Returns the error message.
        //
        // Returns:
        //     The error text.
        public string Message { get; }
        //
        // Summary:
        //     List of file names that failed to upload.
        //
        // Returns:
        //     A string array containing the names of the files that failed to upload.
        public string[] FileNames { get; }
        //
        // Summary:
        //     List of file sizes that failed to upload.
        //
        // Returns:
        //     An integer array containing the sizes of the files that failed to upload.
        public long[] FileSizes { get; }
    }

    public class UploaderProgressEventArgs : EventArgs
    {
        //
        // Summary:
        //     Initializes a new instance of the Wisej.Web.UploadProgressEventArgs class with
        //     the specified error message.
        //
        // Parameters:
        //   loaded:
        //     The number of bytes transferred. This doesn't include headers and other overhead,
        //     but only the content itself.
        //
        //   total:
        //     The total number of bytes of content that will be transferred. If the total size
        //     is unknown, this value is zero.
        public UploaderProgressEventArgs(long loaded, long total, int transferRate)
        {
            this.Loaded = loaded;
            this.Total = total;
            this.TransferRate = transferRate;
        }

        //
        // Summary:
        //     The number of bytes transferred.
        public long Loaded { get; }
        //
        // Summary:
        //     The total number of bytes of content that will be transferred.
        public long Total { get; }
        public int TransferRate { get; }
    }

    public class UploaderFile : IDisposable
    {
        private Stream inputStream;
        public FileInfo File { get; set; }
        public string Name
        {
            get
            {
                if (File == null)
                {
                    return Path.GetFileName(FilePath);
                }
                return File.Name;
            }
        }
        public string FileName
        {
            get
            {
                if (File == null)
                {
                    return Path.GetFileName(FilePath);
                }
                return File.FullName;
            }
        }
        public long Size
        {
            get
            {
                if (File == null)
                {
                    var fi = new FileInfo(FilePath);
                    return fi.Length;
                }
                return File.Length;
            }
        }
        public Stream InputStream
        {
            get
            {
                if (inputStream == null)
                {
                    if (System.IO.File.Exists(FilePath))
                    {
                        inputStream = new FileStream(FilePath, FileMode.Open, FileAccess.Read);
                    }
                    else
                    {
                        return null;
                    }
                }
                return inputStream;
            }
        }

        public string FilePath { get; set; }

        public void SaveAs(string fileName)
        {
            try
            {
                if (File == null)
                {
                    if (System.IO.File.Exists(fileName))
                        System.IO.File.Delete(fileName);

                    System.IO.File.Move(FilePath, fileName);
                }
                else
                {
                    File.MoveTo(fileName);
                }
            }
            catch
            {
                AlertBox.Show("File cannot be saved on server!");
            }
        }

        public void Delete()
        {
            if (string.IsNullOrEmpty(FilePath))
            {
                return;
            }

            Dispose();

            if (System.IO.File.Exists(FilePath))
            {
                System.IO.File.Delete(FilePath);
            }
        }

        public void Dispose()
        {
            inputStream?.Dispose();
            inputStream = null;
        }
    }

    public class SpeedMonitor
    {
        private const int DefaultAveragePeriod = 12;

        private long total;
        private int speed;
        private int[] speeds;
        private int speedsIndex;
        private DateTime lastUpdated;
        private long tempRecvCount;


        public int Rate
        {
            get { return this.speed; }
        }

        public long Total
        {
            get { return this.total; }
        }


        public SpeedMonitor()
            : this(DefaultAveragePeriod)
        {

        }

        public SpeedMonitor(int averagingPeriod)
        {
            if (averagingPeriod < 0)
                throw new ArgumentOutOfRangeException("averagingPeriod");

            this.lastUpdated = DateTime.UtcNow;
            this.speeds = new int[Math.Max(1, averagingPeriod)];
            this.speedsIndex = -speeds.Length;
        }


        public void AddDelta(int speed)
        {
            this.total += speed;
            this.tempRecvCount += speed;
        }

        public void AddDelta(long speed)
        {
            this.total += speed;
            this.tempRecvCount += speed;
        }

        public void Reset()
        {
            this.total = 0;
            this.speed = 0;
            this.tempRecvCount = 0;
            this.lastUpdated = DateTime.UtcNow;
            this.speedsIndex = -speeds.Length;
        }

        private void TimePeriodPassed(int difference)
        {
            int currSpeed = (int)(tempRecvCount * 1000 / difference);
            tempRecvCount = 0;

            int speedsCount;
            if (speedsIndex < 0)
            {
                //speeds array hasn't been filled yet

                int idx = speeds.Length + speedsIndex;

                speeds[idx] = currSpeed;
                speedsCount = idx + 1;

                speedsIndex++;
            }
            else
            {
                //speeds array is full, keep wrapping around overwriting the oldest value
                speeds[speedsIndex] = currSpeed;
                speedsCount = speeds.Length;

                speedsIndex = (speedsIndex + 1) % speeds.Length;
            }

            int total = speeds[0];
            for (int i = 1; i < speedsCount; i++)
                total += speeds[i];

            speed = total / speedsCount;
        }


        public void Tick()
        {
            DateTime old = lastUpdated;
            DateTime now = DateTime.UtcNow;
            int difference = (int)(now - old).TotalMilliseconds;

            if (difference > 800)
            {
                lastUpdated = now;
                TimePeriodPassed(difference);
            }
        }

        // Used purely for unit testing purposes.
        internal void Tick(int difference)
        {
            lastUpdated = DateTime.UtcNow;
            TimePeriodPassed(difference);
        }
    }
}
